/*
 * milter-spamc.c
 *
 * Copyright 2003 by Anthony Howe. All rights reserved.
 *
 * The following should be added to the sendmail.mc file:
 *
 *	INPUT_MAIL_FILTER(
 *		`milter-spamc',
 *		`S=unix:/var/lib/milter-spamc/socket, T=S:30s;R:3m'
 *	)dnl
 */

/*
 * Define NDEBUG for production, undefine for testing (especially
 * on a system with an existing SpamAssassin milter).
 */
#define NDEBUG

#ifndef RUN_AS_USER
#define RUN_AS_USER			"smmsp"
#endif

#ifndef RUN_AS_GROUP
#define RUN_AS_GROUP			"smmsp"
#endif

#ifndef DEFAULT_SPAMD_HOST
#define DEFAULT_SPAMD_HOST		"127.0.0.1"
#endif

#ifndef DEFAULT_SPAMD_PORT
#define DEFAULT_SPAMD_PORT		783
#endif

#ifndef DEFAULT_SOCKET_TIMEOUT
#define DEFAULT_SOCKET_TIMEOUT		120
#endif

#ifndef DEFAULT_HEADER_BUFFER_SIZE
#define DEFAULT_HEADER_BUFFER_SIZE	8192
#endif

#ifndef DEFAULT_SPAM_SUBJECT_TAG
#define DEFAULT_SPAM_SUBJECT_TAG	"[SPAM]"
#endif

#ifndef DEFAULT_SPAM_LEVEL_CHARACTER
#define DEFAULT_SPAM_LEVEL_CHARACTER	'x'
#endif

/*
 * Choose between CHECK, REPORT_IFSPAM, and REPORT.  
 * PROCESS and SYMBOLS not supported.
 */
#ifndef DEFAULT_SPAMC_COMMAND
#define DEFAULT_SPAMC_COMMAND		"REPORT"
#endif

#ifndef STATEDIR
#define STATEDIR			"/var/lib/milter-spamc"
#endif

/*
 * Allow time for smfi_stop() to cleanly shutdown the milter and exit
 * after a critical error, otherwise we fallback on abort().
 */
#ifndef DELAY_BEFORE_ABORT
#define DELAY_BEFORE_ABORT		60
#endif

/***********************************************************************
 *** No configuration below this point.
 ***********************************************************************/

#include "milter-spamc.h"

/* Re-assert this macro just in case. May cause a compiler warning. */
#define _REENTRANT	1

#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <sys/stat.h>
#include <netdb.h>

#include <libmilter/mfapi.h>


#include <com/snert/lib/version.h>
#include <com/snert/lib/io/Socket.h>
#include <com/snert/lib/util/Buf.h>
#include <com/snert/lib/util/Text.h>
#include <com/snert/lib/util/Vector.h>
#include <com/snert/lib/util/getopt.h>

#if LIBSNERT_MAJOR < 1 || LIBSNERT_MINOR < 20
# error "LibSnert/1.20 or better is required"
#endif

/***********************************************************************
 *** Global Variables & Constants
 ***********************************************************************/

#ifndef MILTER_CHUNK_SIZE
#define MILTER_CHUNK_SIZE	65535
#endif

/* See RFC 2821 section 4.5.3.1 Size limits and minimums */
#define SMTP_LINE_SIZE		1000
#define DOMAIN_LENGTH		255
#define LOCAL_PART_LENGTH	128
#define EMAIL_LENGTH		(LOCAL_PART_LENGTH+1+DOMAIN_LENGTH)
#define IPV6_LENGTH		((128/16)*4+(128/16)-1)

#define PARSE_PATH		"<%384[^>]>%n"		/* EMAIL_LENGTH = 384 */
#define PARSE_MAILBOX		"%128[^@]%n@%255s%n"	/* LOCAL_PART_LENGTH & DOMAIN_LENGTH */
#define FORMAT_TIMESTAMP	"; %a, %d %b %Y %H:%M:%S %z"

#define SMTP_OK			250
#define SMTP_IS_OK(x)		(200 <= (x) && (x) < 300)
#define SMTP_IS_DEFER(x)	(300 <= (x) && (x) < 400)
#define SMTP_IS_TEMP(x)		(400 <= (x) && (x) < 500)
#define SMTP_IS_PERM(x)		(500 <= (x) && (x) < 600)

typedef struct {
	SMFICTX *ctx;				/* pointer to this */
	Socket server;				/* per message */
	Vector rcpts;				/* per message */
	Buf headers;				/* per message */
	int byPass;				/* per message */
	int isSent;				/* per message */
	int isSpamRcpt;				/* per message */
	char *qid;				/* per message */
	char *localRcpt;			/* per message */
	char line[SMTP_LINE_SIZE+1];		/* general purpose */
	char subject[SMTP_LINE_SIZE+1];		/* subject header */
	char replyLine[SMTP_LINE_SIZE+1];	/* setReply(), vSetReply() */
} *workspace;

#define DEBUG_ALL		(~0L)
#define DEBUG_INFO		1
#define DEBUG_HOOKS		2
#define DEBUG_SPAMD		4
#define DEBUG_SOCKETS		8

static long debug;
static long socketTimeout;

static char *spamcCommand;
static int alwaysAddReport;
static int forceSpamdUser;
static int redirectSpam;
static int spamdPort;
static int appendScoreToSubjectTag;
static char *spamdHost;
static char *spamdUser;
static char *copyAllTo;
static char *copyHamTo;
static char *copySpamTo;
static char *subjectTag;
static char *subjectScoreTag;
static long rejectExcessScore;
static long discardExcessScore;

static char hostaddr[IPV6_LENGTH+1];
static char hostname[DOMAIN_LENGTH+1];

#ifdef SMFIS_REJECT_AND_DELIVER
static long rejectAndDeliverExcessScore;
#endif

#ifdef SMFIS_REJECT_AND_DELIVER
static char getoptString[] = "ACSa:b:B:c:d:h:r:R:s:t:u:U:v:H:P:?";
#else
static char getoptString[] = "ACSa:b:B:c:d:h:r:s:t:u:U:v:H:P:?";
#endif

static char usageMessage[] =
"\n"
"usage:\n"
"\n"
#ifdef SMFIS_REJECT_AND_DELIVER
"milter-spamc [-A|-C][-a email][-b email|-B email][-c email][-d excess]\n"
"         [-r excess][-R excess][-s tag][-t timeout][-u user|-U user]\n"
"         [-v debug][-H host][-P port] socket\n"
#else
"milter-spamc [-A|-C][-a email][-b email|-B email][-c email][-d excess]\n"
"        [-r excess][-s tag][-S][-t timeout][-u user|-U user][-v debug]\n"
"        [-H host][-P port] socket\n"
#endif
"\n"
"-a email\tcopy all mail to email address\n"
"-A\t\talways add the X-Spam-Report: header report of rules triggered\n"
"-b email\tcopy junk mail to email address\n"
"-B email\tredirect junk mail to email address\n"
"-c email\tcopy clean mail to email address\n"
"-C\t\tcheck for spam only, disable X-Spam-Report: header completely\n"
"-d excess\tdiscard spam messages that exceed the threshold by N\n"
"-h\t\tthis usage summary\n"
"-H host\t\tspamd host name or IP\n"
"-P port\t\tspamd port address, -1 to disable spamd filtering\n"
"-r excess\treject spam that exceeds the threshold by N, overrides delivery\n"
#ifdef SMFIS_REJECT_AND_DELIVER
"-R excess\treject spam that exceeds the threshold by N, allow delivery\n"
#endif
"-s tag\t\tspam subject tag\n"
"-S\t\tappend the spam score to the spam subject tag\n"
"-t timeout\tspamd I/O timeout in seconds, 0 indefinite\n"
"-u user\t\tdefault user account to process messages with\n"
"-U user\t\tforce message processing with this user account\n"
"-v debug\tthe debug bit mask\n"
"\n"
MILTER_VERSION " " MILTER_COPYRIGHT "\n"
"\n"
;

#ifdef NDEBUG
/* Headers as they should appear when in production. */

#  define X_SCANNED_BY			"X-Scanned-By"
#  define X_SPAM_FLAG			"X-Spam-Flag"
#  define X_SPAM_LEVEL			"X-Spam-Level"
#  define X_SPAM_REPORT			"X-Spam-Report"
#  define X_SPAM_STATUS			"X-Spam-Status"
#  define X_ORIGINAL_RECIPIENT		"X-Original-Recipient"

#else

/* Headers as they should appear when testing, especially in
 * conjunction with other SpamAssassin milters that handle
 * the same set of headers. This allows me to test and see
 * the headers added by my milter and compare with those of
 * others.
 */

#  define X_SCANNED_BY			"X-" MILTER_NAME "-Scanned-By"
#  define X_SPAM_FLAG			"X-" MILTER_NAME "-Flag"
#  define X_SPAM_LEVEL			"X-" MILTER_NAME "-Level"
#  define X_SPAM_REPORT			"X-" MILTER_NAME "-Report"
#  define X_SPAM_STATUS			"X-" MILTER_NAME "-Status"
#  define X_ORIGINAL_RECIPIENT		"X-" MILTER_NAME "-Original-Recipient"

#  define DEFAULT_SPAM_SUBJECT_TAG	"[" MILTER_NAME "]"

#endif

/***********************************************************************
 *** Support routines.
 ***********************************************************************/

#ifdef GETHOSTBYNAME_MUTEX_WRAPPER
static void *theMutex;

void
lockTheMutex(void)
{
	if (MutexLock(theMutex)) {
		syslog(LOG_CRIT, "stopping milter, failed to lock mutex");
		die();
	}
}

void
unlockTheMutex(void)
{
	if (MutexUnlock(theMutex)) {
		syslog(LOG_CRIT, "stopping milter, failed to unlock mutex");
		die();
	}
}
#endif

#ifndef HAVE_INET_NTOP
void
getIp4Octets(unsigned long ip, int *a, int *b, int *c, int *d)
{
	*a = (ip >> 24) & 0xff;
	*b = (ip >> 16) & 0xff;
	*c = (ip >> 8 ) & 0xff;
	*d = (ip      ) & 0xff;
}

char *
ip4ToString(unsigned long ip)
{
	char *s;
	int a, b, c, d;

	if ((s = malloc(16)) != (char *) 0) {
		getIp4Octets(ip, &a, &b, &c, &d);
		(void) snprintf(s, 16, "%u.%u.%u.%u", a,b,c,d);
	}

	return s;
}

const char *
inet_ntop(int af, const void *in_addr, char *buffer, size_t size)
{
	int a, b, c, d;

	getIp4Octets(ntohl(((struct in_addr *) in_addr)->s_addr), &a, &b, &c, &d);
	(void) snprintf(buffer, size, "%u.%u.%u.%u", a,b,c,d);

	return buffer;
}
#endif

void
die()
{
	smfi_stop();
	sleep(DELAY_BEFORE_ABORT);

	/* We should not normally reach this if the smfi_stop()
	 * kills this thread and the milter in a reasonable
	 * amount of time.
	 */
	abort();
}

sfsistat
nullWorkspaceError(SMFICTX *ctx)
{
	/* Without the workspace we can't do a hell of a lot. */
	(void) smfi_setreply(ctx, "421", "4.3.2", "close connection");
	syslog(LOG_ERR, "internal error: null workspace");
	return SMFIS_TEMPFAIL;
}

sfsistat
vSetReply(workspace data, int code, char *ecode, char *fmt, va_list args)
{
	sfsistat rc;
	char codeString[4], extended[6];

	if (SMTP_IS_PERM(code))
		rc = SMFIS_REJECT;
	else if (SMTP_IS_TEMP(code))
		rc = SMFIS_TEMPFAIL;
	else
		rc = SMFIS_CONTINUE;

	(void) snprintf(codeString, sizeof(codeString), "%d", code);

	if (ecode == (char *) 0) {
		(void) snprintf(extended, sizeof(extended), "%c.7.1", *codeString);
		ecode = extended;
	}

	/* Build reply message. */
	(void) vsnprintf(data->replyLine, SMTP_LINE_SIZE, fmt, args);

	/* Write our error to the log, particularly important for -g,
	 * for debugging without having enable full debug messages.
	 */
	syslog(LOG_ERR, "%s: reply %s %s %s", data->qid, codeString, ecode, data->replyLine);

	/* Tell sendmail what error to report to the sender. */
	(void) smfi_setreply(data->ctx, codeString, ecode, data->replyLine);

	return rc;
}

sfsistat
setReply(workspace data, int code, char *ecode, char *fmt, ...)
{
	sfsistat rc;

	va_list args;
	va_start(args, fmt);

	rc = vSetReply(data, code, ecode, fmt, args);

	va_end(args);

	return rc;
}

/*
 * Write buffer, logging only errors.
 */
int
writebuffer(workspace data, char *buf, long size)
{
	int error;
	int rc = 0;
	long nbytes, length;

	/* Do not block on sending to the spamd server just yet. */
	(void) SocketSetNonBlocking(data->server, 1);

	for (length = 0; length < size; length += nbytes) {
		if ((nbytes = SocketWrite(data->server, buf + length, size - length)) < 0) {
			error = SocketGetError(data->server);
			syslog(LOG_ERR, "%s: SocketWrite() error: %s (%d)", data->qid, strerror(error), error);
			goto error0;
		}

		/* Now wait for the output to be sent to the spamd server. */
		if (SocketWaitForOutput(data->server, socketTimeout) != 0) {
			syslog(LOG_ERR, "%s: timeout before output sent to SPAMD server", data->qid);
			goto error0;
		}
	}

	rc = 1;
error0:
	(void) SocketSetNonBlocking(data->server, 0);

	return rc;
}

/*
 * Log line to be sent then write line.
 */
int
writeline(workspace data, char *line, long size)
{
	if (debug & DEBUG_SPAMD)
		syslog(LOG_DEBUG, "%s: > %s", data->qid, line);

	return writebuffer(data, line, size < 0 ? strlen(line) : size);
}

/*
 * Log amount of data sent then write data.
 */
int
writechunk(workspace data, char *chunk, long size)
{
	if (debug & DEBUG_SPAMD)
		syslog(LOG_DEBUG, "%s: > [%ld bytes of data]", data->qid, size);

	return writebuffer(data, chunk, size);
}

int
readline(workspace data, char *line, int linesize)
{
	long length;
	int rc, error;

	rc = -1;

	if (linesize < sizeof (long)) {
		syslog(LOG_CRIT, "%s: line buffer too small (%d)", data->qid, linesize);
		die();
	}

	error = SocketWaitForInput(data->server, socketTimeout);

	if (error == ETIMEDOUT) {
		syslog(LOG_ERR, "%s: timeout before input from SPAMD server", data->qid);
		goto error0;
	}

	if (error != 0) {
		syslog(LOG_ERR, "%s: input error from SPAMD server: %s (%d)", data->qid, strerror(error), error);
		goto error0;
	}

	length = SocketReadLine(data->server, line, linesize);

	if ((error = SocketGetError(data->server)) != 0 || length == -1) {
		syslog(LOG_ERR, "%s: read error from SPAMD server: %s (%d)", data->qid, strerror(error), error);
		goto error0;
	}

	if (debug & DEBUG_SPAMD)
		syslog(LOG_DEBUG, "%s: < %s", data->qid, line);

	rc = length;
error0:
	return rc;
}

/***********************************************************************
 *** Handlers
 ***********************************************************************/

/*
 * Open and allocate per-connection resources.
 */
sfsistat
filterOpen(SMFICTX *ctx, char *hostname, _SOCK_ADDR *hostaddr)
{
	int rc;
	workspace data;

	if (debug & DEBUG_HOOKS)
		syslog(LOG_DEBUG, "create per-connection resources...");

	if ((data = calloc(1, sizeof *data)) == (workspace) 0)
		goto error0;

	if ((data->headers = BufCreate(DEFAULT_HEADER_BUFFER_SIZE)) == (Buf) 0)
		goto error1;

	if ((data->rcpts = VectorCreate(10)) == (Vector) 0)
		goto error2;

	VectorOnRemove(data->rcpts, VectorRemoveTrivial, (void *) 0);

	if (smfi_setpriv(ctx, (void *) data) == MI_FAILURE)
		goto error3;

	/* Remember the context we belong to. Useful when passing around
	 * numerous arguements.
	 */
	data->ctx = ctx;
	data->qid = "NOQUEUE";

	return SMFIS_CONTINUE;
error3:
	VectorDestroy(data->rcpts);
error2:
	BufDestroy(data->headers);
error1:
	free(data);
error0:
	return SMFIS_REJECT;
}

sfsistat
filterMail(SMFICTX *ctx, char **args)
{
	sfsistat rc;
	workspace data;
	char *auth_authen;

	if ((data = (workspace) smfi_getpriv(ctx)) == (workspace) 0)
		return nullWorkspaceError(ctx);

	/* Reset per-message variables, since this function can be
	 * called several times per message or connection, ie.
	 * MAIL-RSET-MAIL or MAIL-RCPT-DATA-MAIL sequences.
	 */

	VectorRemove(data->rcpts, 0, VectorLength(data->rcpts));
	BufSetLength(data->headers, 0);

	if (data->localRcpt != spamdUser)
		free(data->localRcpt);

	data->qid = smfi_getsymval(ctx, "i");
	data->localRcpt = (char *) 0;
	data->byPass = spamdPort < 0;
	data->isSpamRcpt = 0;
	data->isSent = 0;

	/* Make sure its closed before we start the next message. */
	SocketClose(data->server);
	data->server = (Socket) 0;

	/* We assume that an authenticated user is trusted not to send spam. */
	auth_authen = smfi_getsymval(data->ctx, "{auth_authen}");
	if (auth_authen != (char *) 0) {
		if (debug & (DEBUG_INFO|DEBUG_HOOKS))
			syslog(LOG_INFO, "%s: authenticated \"%s\", skipping", data->qid, auth_authen);
		data->byPass = 1;
	}

	return SMFIS_CONTINUE;
}

sfsistat
filterRcpt(SMFICTX *ctx, char **args)
{
	sfsistat rc;
	workspace data;
	char *rcpt, *rcpt_addr, *rcpt_host, *rcpt_mailer;

	if ((data = (workspace) smfi_getpriv(ctx)) == (workspace) 0)
		return nullWorkspaceError(ctx);

	if (data->byPass)
		return SMFIS_CONTINUE;

	rcpt_addr = smfi_getsymval(data->ctx, "{rcpt_addr}");
	rcpt_host = smfi_getsymval(data->ctx, "{rcpt_host}");
	rcpt_mailer = smfi_getsymval(data->ctx, "{rcpt_mailer}");

	/* Remember last local user. Ideally we should also attempt to
	 * expand mail aliases, but that means lots of Berkeley DB issues,
	 * so for teh time being we'll ignore it and hope its not that
	 * important.
	 */
	if (*rcpt_host == '\0' && getpwnam(rcpt_addr) != (struct passwd *) 0) {
		/* Free NULL or a previous local recipient. */
		free(data->localRcpt);

		if (debug & DEBUG_HOOKS)
			syslog(LOG_DEBUG, "%s: address %s has an account", data->qid, args[0]);

		/* Assign local recipient. Don't worry if NULL, handled in filterBody(). */
		data->localRcpt = TextDup(rcpt_addr);
	}

	/* Add rcpt to list of recipients. */
	if ((rcpt = TextDup(args[0])) == (char *) 0)
		return setReply(data, 452, "4.3.2", "out of memory, cannot duplicate recipient");

	if (VectorAdd(data->rcpts, rcpt)) {
		free(rcpt);
		return setReply(data, 452, "4.3.2", "out of memory, cannot add recipient to list");
	}

	if (debug & DEBUG_HOOKS)
		syslog(
			LOG_DEBUG, "%s: RCPT='%s' rcpt_addr='%s' rcpt_host='%s' rcpt_mailer='%s'",
			data->qid, rcpt, rcpt_addr, rcpt_host, rcpt_mailer
		);

	return SMFIS_CONTINUE;
}

sfsistat
filterHeader(SMFICTX *ctx, char *name, char *value)
{
	workspace data;

	if ((data = (workspace) smfi_getpriv(ctx)) == (workspace) 0)
		return nullWorkspaceError(ctx);

	if (data->byPass)
		return SMFIS_CONTINUE;

	if (TextInsensitiveCompare(name, "Subject") == 0)
		(void) strncpy(data->subject, value, SMTP_LINE_SIZE - strlen(subjectTag) - 1);

	(void) BufAddString(data->headers, name);
	(void) BufAddBytes(data->headers, ": ", 0, 2);
	(void) BufAddString(data->headers, value);
	(void) BufAddBytes(data->headers, "\r\n", 0, 2);

	return SMFIS_CONTINUE;
}

sfsistat
filterEndHeaders(SMFICTX *ctx)
{
	workspace data;

	if ((data = (workspace) smfi_getpriv(ctx)) == (workspace) 0)
		return nullWorkspaceError(ctx);

	if (data->byPass)
		return SMFIS_CONTINUE;

	(void) BufAddBytes(data->headers, "\r\n", 0, 2);

	return SMFIS_CONTINUE;
}

sfsistat
filterBody(SMFICTX *ctx, unsigned char *chunk, size_t size)
{
	long length;
	workspace data;
	char *msg_size, *if_addr;

	if ((data = (workspace) smfi_getpriv(ctx)) == (workspace) 0)
		return nullWorkspaceError(ctx);

	if (data->byPass || data->isSent)
		return SMFIS_CONTINUE;

	if (smfi_getsymval(ctx, "{if_addr}") == (char *) 0) {
		syslog(LOG_INFO, "%s: loopback address, skipping scan", data->qid);
		return SMFIS_CONTINUE;
	}

 	/* If the redirectSpamTo address is the only address, skip filtering. */
 	if (data->isSpamRcpt && VectorLength(data->rcpts) == 0)
		return SMFIS_CONTINUE;

	/* Send only the headers and the first chunk to spamd. */
	if (size == 0) {
		syslog(LOG_INFO, "%s: empty message body, skipping scan");
		return SMFIS_CONTINUE;
	}

 	/* Only process the headers and first body chunk. */
 	data->isSent = 1;

#ifdef GETHOSTBYNAME_MUTEX_WRAPPER
/* SocketOpenTcpClient() and SocketOpenTcpClientWait() uses
 * gethostbyname(), which is not thread safe and there is no
 * gethostbyname_r() on the BSD based systems.
 */
		lockTheMutex();
#endif

	data->server = SocketOpenTcpClientWait(spamdHost, spamdPort, socketTimeout);

#ifdef GETHOSTBYNAME_MUTEX_WRAPPER
		unlockTheMutex();
#endif
	if (data->server == (Socket) 0) {
		syslog(LOG_ERR, "%s: failed to connect to spamd %s,%d", data->qid, spamdHost, spamdPort);
		goto error0;
	}

	if (debug & DEBUG_HOOKS)
		syslog(LOG_DEBUG, "%s: SPAMD Content-Length=%ld", data->qid, BufLength(data->headers) + size);

	/* Ask for a CHECK or REPORT only and we'll handle adding the X-Spam-* headers. */
	length = snprintf(
		data->line, SMTP_LINE_SIZE, "%s SPAMC/1.2\r\nContent-Length: %ld\r\n",
		spamcCommand, BufLength(data->headers) + size
	);
	if (!writeline(data, data->line, length))
		goto error1;

	/* If the spamd user is defined, then we'll use it if forced or
	 * multiple recipients given. Otherwise fall back on the last
	 * defined local recipient if any.
	 */
	if (spamdUser != (char *) 0 && (forceSpamdUser || 1 < VectorLength(data->rcpts))) {
		/* Free NULL or a previous local user. */
		free(data->localRcpt);
		data->localRcpt = (char *) 0;
	}

	/* When no local recipient defined, then fallback on the default
	 *spamd user, which might be undefined.
	 */
	if (data->localRcpt == (char *) 0)
		data->localRcpt = spamdUser;

	if (data->localRcpt != (char *) 0) {
		if (debug & DEBUG_HOOKS)
			syslog(LOG_DEBUG, "%s: SPAMD User=%s", data->qid, data->localRcpt);
		length = snprintf(data->line, SMTP_LINE_SIZE, "User: %s\r\n", data->localRcpt);
		if (!writeline(data, data->line, length))
			goto error1;
	}

	if (!writeline(data, "\r\n", 2))
		goto error1;

	if (!writechunk(data, BufBytes(data->headers), BufLength(data->headers)))
		goto error1;

	if (!writechunk(data, chunk, size))
		goto error1;

	/* Signal EOF to SPAMD so that it can begin processing. */
	SocketShutdownNow(data->server, SHUT_WR);

	return SMFIS_CONTINUE;
error1:
	SocketClose(data->server);
	data->server = (Socket) 0;
error0:
	return SMFIS_CONTINUE;
}

sfsistat
filterEndMessage(SMFICTX *ctx)
{
	char *rcpt;
	time_t now;
	workspace data;
	struct tm timeNow;
	float score, threshold;
	int i, isSpam, length, iscore;

	if ((data = (workspace) smfi_getpriv(ctx)) == (workspace) 0)
		return nullWorkspaceError(ctx);

	/* Send a copy of all mail to the archive address. */
	if (copyAllTo != (char *) 0 && smfi_addrcpt(ctx, copyAllTo) == MI_FAILURE)
		syslog(LOG_ERR, "%s: copy-all-to <%s> failed", data->qid, copyAllTo);

	/* Skip the remainder of this function if there was an
	 * earlier error or if we did not start a spamd session.
	 */
	if (data->byPass || data->server == (Socket) 0)
		return SMFIS_CONTINUE;

	(void) time(&now);
	length = snprintf(data->line, SMTP_LINE_SIZE, MILTER_VERSION " (%s [%s])", hostname, hostaddr);			
	if (localtime_r((const time_t *) &now, &timeNow) != (struct tm *) 0)
		(void) strftime(data->line + length, SMTP_LINE_SIZE - length, FORMAT_TIMESTAMP, &timeNow);			
	(void) smfi_addheader(ctx, X_SCANNED_BY, data->line);

	/* Fetch spamd result and add it to the headers. */
	if (readline(data, data->line, SMTP_LINE_SIZE) <= 0) {
		syslog(LOG_ERR, "%s: SPAMD status line failure", data->qid);
		return SMFIS_CONTINUE;
	}

	if (strncmp(data->line, "SPAMD/1.1 0 EX_OK", sizeof ("SPAMD/1.1 0 EX_OK")) != 0) {
		syslog(LOG_ERR, "%s: SPAMD failure: %s", data->qid, data->line);
		return SMFIS_CONTINUE;
	}

	/* Set to -1 for undefined, since TextInsensitiveStartsWith() 
	 * returns 0 on no match or a postive length for a match.
	 */
	isSpam = -1;
	
	while (0 < readline(data, data->line, SMTP_LINE_SIZE)) {
		if (TextInsensitiveStartsWith(data->line, "Spam:") != 0) {
			/* And the verdict is? */
			isSpam = TextInsensitiveStartsWith(data->line, "Spam: True");
			(void) smfi_addheader(ctx, X_SPAM_FLAG, isSpam ? "YES" : "NO");

			/* Extract the score. */
			if (sscanf(data->line, "%*[^;]; %f / %f", &score, &threshold) != 2) {
				syslog(LOG_ERR, "%s: Spam: header parse error: %s", data->qid, data->line);
				return SMFIS_CONTINUE;
			}
		}
		
		/* Ignore any other headers. */
	}
	
	if (isSpam == -1) {
		syslog(LOG_ERR, "%s: missing Spam: result header", data->qid);
		return SMFIS_CONTINUE;
	}

	/* Add a pretty header. I think this one is useless. */
	(void) sprintf(data->line, "%s, hits=%.2f required=%.2f", isSpam ? "YES" : "NO", score, threshold);
	(void) smfi_addheader(ctx, X_SPAM_STATUS, data->line);

	/* Add another header that can be parsed with by a regular expression.
	 * This one is useful for filtering different levels of spam. Note that
	 * we initially use a non-regular-expression character for writing the
	 * level.
	 */
	if (SMTP_LINE_SIZE <= score)
		/* X-Spam-Level: xxx...x\r\n\0 */
		score = SMTP_LINE_SIZE - sizeof (X_SPAM_LEVEL) - 5;

	iscore = (int) score;
	for (i = 0; i < iscore; i++)
		data->line[i] = DEFAULT_SPAM_LEVEL_CHARACTER;
	data->line[i] = '\0';

	(void) smfi_addheader(ctx, X_SPAM_LEVEL, data->line);

	/* REUSE this buffer to collect the report. */
	BufSetLength(data->headers, 0);

	/* SpamAssassin 2.5x prefixed a redundant score line for a report. */
	length = readline(data, data->line, SMTP_LINE_SIZE);
	if (strspn(data->line, "1234567890-./") != length) {
		/* Remove trailing whitespace. */
		while (0 < length-- && isspace(data->line[length]))
			data->line[length] = '\0';

		/* Only add non-empty lines. */
		if (data->line[0] != '\0') {
			BufAddString(data->headers, data->line);
			BufAddString(data->headers, "\n  ");
		}
	}

	/* Collect report lines until read error or EOF. */
	while (0 <= (length = readline(data, data->line, SMTP_LINE_SIZE))) {
		/* Remove trailing whitespace. */
		while (0 < length-- && isspace(data->line[length]))
			data->line[length] = '\0';

		if (length <= 0) {
			/* Replace a empty line with something that looks blank. */
			BufAddString(data->headers, "____\n  ");
		} else {
			BufAddString(data->headers, data->line);
			BufAddString(data->headers, "\n  ");
		}
	}

	/* Done with spamd session. */
	SocketClose(data->server);
	data->server = (Socket) 0;

	/* Remove our trailing whitespace "\r\n  ". */
	BufSetLength(data->headers, BufLength(data->headers)-3);

	if (alwaysAddReport)
		/* Add the report header. */
		(void) smfi_addheader(ctx, X_SPAM_REPORT, BufBytes(data->headers));

	if (isSpam) {
		if (!alwaysAddReport)
			/* Add the report header. */
			(void) smfi_addheader(ctx, X_SPAM_REPORT, BufBytes(data->headers));

		/* Tag the subject if not already. */
		if (TextInsensitiveStartsWith(data->subject, subjectTag) == 0) {
			if (appendScoreToSubjectTag)			
				(void) snprintf(data->line, sizeof data->line, "%s (%.2f/%.2f) %s", subjectTag, score, threshold, data->subject);
			else
				(void) snprintf(data->line, sizeof data->line, "%s %s", subjectTag, data->subject);
			(void) smfi_chgheader(ctx, "Subject", 1, data->line);
		}

		if (copySpamTo != (char *) 0) {
			if (smfi_addrcpt(ctx, copySpamTo) == MI_FAILURE) {
				syslog(LOG_ERR, "%s: copy-spam-to <%s> failed ", data->qid, copySpamTo);
				goto error1;
			}

			if (redirectSpam) {
				/* Remove previous list of recipients from Sendmail's list
				 * and record them in the redirected message in case it was
				 * not spam and needs to be resent later. Sometimes the To:
				 * or Cc: headers are useless ie. undisclosed recipients.
				 */
				for (i = 0; i < VectorLength(data->rcpts); i++) {
					rcpt = VectorGet(data->rcpts, i);

					/* Don't worry about rcpt == NULL since smfi_addheader()
					 * and smfi_delrcpt() already test for it.
					 */

					if (smfi_addheader(ctx, X_ORIGINAL_RECIPIENT, rcpt) == MI_SUCCESS)
						(void) smfi_delrcpt(ctx, rcpt);
				}
			}
#ifndef NDEBUG
			else {
				/* Save copy of recipient list to headers for debugging. */
				for (i = 0; i < VectorLength(data->rcpts); i++) {
					rcpt = VectorGet(data->rcpts, i);
					(void) smfi_addheader(ctx, X_ORIGINAL_RECIPIENT, rcpt);
				}
			}
#endif
		}
	} else {
		/* Send a copy of clean mail to an archive address. */
		if (copyHamTo != (char *) 0 && smfi_addrcpt(ctx, copyHamTo) == MI_FAILURE)
			syslog(LOG_ERR, "%s: copy-ham-to <%s> failed ", data->qid, copyHamTo);
	}
error1:
	syslog(LOG_INFO, "%s: spam=%s score=%.2f required=%.2f reject-excess=%ld discard-excess=%ld", data->qid, isSpam ? "YES" : "NO", score, threshold, rejectExcessScore, discardExcessScore);		
error0:
	if (isSpam) {
		(void) setReply(data, 550, "5.7.1", "message was identified as junk mail");

		if (threshold + rejectExcessScore <= score)
			return SMFIS_REJECT;
		if (threshold + discardExcessScore <= score)
			return SMFIS_DISCARD;
#ifdef SMFIS_REJECT_AND_DELIVER
		if (threshold + rejectAndDeliverExcessScore <= score)
			return SMFIS_REJECT_AND_DELIVER;
#endif
	}

	return SMFIS_CONTINUE;
}

/*
 * Close and release per-connection resources.
 */
sfsistat
filterClose(SMFICTX *ctx)
{
	workspace data = (workspace) smfi_getpriv(ctx);

	if (data != (workspace) 0) {
		/* We have to do this, otherwise we get a warning. */
		(void) smfi_setpriv(ctx, (void *) 0);

		if (data->localRcpt != spamdUser)
			free(data->localRcpt);

		VectorDestroy(data->rcpts);
		BufDestroy(data->headers);
		SocketClose(data->server);
		free(data);

		if (debug & DEBUG_HOOKS)
			syslog(LOG_DEBUG, "destroyed per-connection resources");
	}

	return SMFIS_CONTINUE;
}

/***********************************************************************
 ***  Milter Definition Block
 ***********************************************************************/

struct smfiDesc smfilter = {
	MILTER_NAME,		/* filter name */
	SMFI_VERSION,		/* version code -- do not change */
	SMFIF_ADDHDRS|SMFIF_CHGHDRS|SMFIF_ADDRCPT|SMFIF_DELRCPT,	/* flags */
	filterOpen,		/* connection info filter */
	NULL,			/* SMTP HELO command filter */
	filterMail,		/* envelope sender filter */
	filterRcpt,		/* envelope recipient filter */
	filterHeader,		/* header filter */
	filterEndHeaders,	/* end of header */
	filterBody,		/* body block filter */
	filterEndMessage,	/* end of message */
	NULL,			/* message aborted */
	filterClose,		/* connection cleanup */
};

/***********************************************************************
 *** Startup
 ***********************************************************************/

int
main(int argc, char **argv)
{
	int ch, rc;
	struct group *gr;
	struct passwd *pw;
	struct hostent *host;
	char *milterSocket = (char *) 0;

	openlog(MILTER_NAME, LOG_PID, LOG_MAIL);

	syslog(LOG_DEBUG, "process uid=%d gid=%d before", getuid(), getgid());

	if ((pw = getpwnam(RUN_AS_USER)) == (struct passwd *) 0) {
		(void) fprintf(stderr, "user \"%s\" not found\n", RUN_AS_USER);
		syslog(LOG_ERR, "user \"%s\" not found", RUN_AS_USER);
		return 1;
	}

	if ((gr = getgrnam(RUN_AS_GROUP)) == (struct group *) 0) {
		(void) fprintf(stderr, "group \"%s\" not found\n", RUN_AS_GROUP);
		syslog(LOG_ERR, "group \"%s\" not found", RUN_AS_GROUP);
		return 1;
	}

	/* Avoid running as root user and/or group. */
	if (getuid() == 0 && pw->pw_uid != 0 && gr->gr_gid != 0) {
		(void) setgid(gr->gr_gid);
		(void) setuid(pw->pw_uid);

		syslog(LOG_DEBUG, "process uid=%d gid=%d after", getuid(), getgid());
	}

	debug = DEBUG_INFO;
	alwaysAddReport = 0;
	copyAllTo = (char *) 0;
	copyHamTo = (char *) 0;
	copySpamTo = (char *) 0;	
	socketTimeout = DEFAULT_SOCKET_TIMEOUT;
	subjectTag = DEFAULT_SPAM_SUBJECT_TAG;
	appendScoreToSubjectTag = 0;
	spamdHost = DEFAULT_SPAMD_HOST;
	spamdPort = DEFAULT_SPAMD_PORT;
	spamdUser = (char *) 0;
	forceSpamdUser = 0;
	redirectSpam = 0;
	spamcCommand = DEFAULT_SPAMC_COMMAND;
	rejectExcessScore = ~(unsigned long)0 >> 1;
	discardExcessScore = ~(unsigned long)0 >> 1;
#ifdef SMFIS_REJECT_AND_DELIVER
	rejectAndDeliverExcessScore = ~(unsigned long)0 >> 1;
#endif

	while ((ch = getopt(argc, argv, getoptString)) != -1) {
		switch (ch) {
		case 'A':
			alwaysAddReport = 1;
			spamcCommand = "REPORT";
			break;
		case 'C':
			alwaysAddReport = 0;
			spamcCommand = "CHECK";
			break;
		case 'a':
			copyAllTo = optarg;
			break;
		case 'b':
			copySpamTo = optarg;
			redirectSpam = 0;
			break;
		case 'B':
			copySpamTo = optarg;
			redirectSpam = 1;
			break;
		case 'c':
			copyHamTo = optarg;
			break;
		case 'd':
			discardExcessScore = strtol(optarg, (char **) 0, 10);
			if (discardExcessScore < 0) {
				(void) fprintf(stderr, "-d value must be positive");
				(void) fprintf(stderr, usageMessage);
			}
			break;
		case 't':
			socketTimeout = strtol(optarg, (char **) 0, 10);
			break;
		case 'r':
			rejectExcessScore = strtol(optarg, (char **) 0, 10);
			if (rejectExcessScore < 0) {
				(void) fprintf(stderr, "-r value must be positive");
				(void) fprintf(stderr, usageMessage);
			}
			break;
#ifdef SMFIS_REJECT_AND_DELIVER
		case 'R':
			rejectAndDeliverExcessScore = strtol(optarg, (char **) 0, 10);
			if (rejectAndDeliverExcessScore < 0) {
				(void) fprintf(stderr, "-R value must be positive");
				(void) fprintf(stderr, usageMessage);
			}
			break;
#endif
		case 's':
			subjectTag = optarg;
			break;
		case 'S':
			appendScoreToSubjectTag = 1;
			break;
		case 'u':
			spamdUser = optarg;
			forceSpamdUser = 0;
			break;
		case 'U':
			spamdUser = optarg;
			forceSpamdUser = 1;
			break;
		case 'v':
			debug = strtol(optarg, (char **) 0, 10);
			break;
		case 'H':
			spamdHost = optarg;
			break;
		case 'P':
			spamdPort = (int) strtol(optarg, (char **) 0, 10);
			break;
		default:
			syslog(LOG_ERR, "unknown command line option -%c", optopt);
			/* FALLTHROUGH */
		case 'h':
		case '?':
			(void) fprintf(stderr, usageMessage);
			return 2;
		}
	}

	socketTimeout *= 1000;

	if (optind + 1 != argc) {
		syslog(LOG_ERR, "missing command line argument");
		(void) fprintf(stderr, usageMessage);
		return 2;
	}

	milterSocket = argv[optind];
	umask(S_IXUSR | S_IRWXG | S_IRWXO);

#ifdef GETHOSTBYNAME_MUTEX_WRAPPER
	if ((theMutex = MutexCreate(THE_MUTEX)) == (void *) 0) {
		syslog(LOG_ERR, "failed to allocate mutex\n");
		return 1;
	}

	if (MutexPermission(theMutex, 0600, pw->pw_uid, gr->gr_gid) < 0) {
		syslog(LOG_ERR, "failed to set mutex permissions\n");
		return 1;
	}
#endif

	if (smfi_setconn(milterSocket) == MI_FAILURE) {
		syslog(LOG_ERR, "smfi_setconn() failed\n");
		return 1;
	}

	if (strncmp(milterSocket, "unix:", 5) == 0) {
		unlink(milterSocket + 5);
	} else if (strncmp(milterSocket, "local:", 6) == 0) {
		unlink(milterSocket + 6);
	}

	if (smfi_register(smfilter) == MI_FAILURE) {
		syslog(LOG_ERR, "smfi_register() failed\n");
		return 1;
	}

	if (debug & DEBUG_SOCKETS)
		SocketSetDebugMask(SOCKET_DEBUG_OPEN_CLOSE|SOCKET_DEBUG_READ|SOCKET_DEBUG_WRITE);

	if (SocketInit()) {
		syslog(LOG_ERR, "SocketInit() error\n");
		return 1;
	}

	gethostname(hostname, sizeof hostname);
	if (debug & DEBUG_HOOKS)
		syslog(LOG_DEBUG, "gethostname() hostname=%s", hostname);

	if ((host = gethostbyname(hostname)) == (struct hostent *) 0) {
		syslog(LOG_ERR, "gethostbyname(%s) failed", hostname);
		return 1;
	}

	if (inet_ntop(host->h_addrtype, host->h_addr_list[0], hostaddr, sizeof hostaddr) == NULL) {
		syslog(LOG_ERR, "inet_ntop() failed");
		return 1;
	}

	/* Don't need these and they just use up resources. */
	fclose(stderr);
	fclose(stdout);
	fclose(stdin);

	/* Display version information in log. */
	syslog(LOG_INFO, MILTER_VERSION " " MILTER_COPYRIGHT);
	syslog(LOG_INFO, LIBSNERT_VERSION " " LIBSNERT_COPYRIGHT);

	return smfi_main();
}
